/*
 *      UTIL.C
 *      UTREE utility functions.
 *      3.01-um klin, Sat Apr 20 11:02:33 1991
 *              klin, Tue Oct 15 14:02:37 1991, Handling of symlinks changed
 *      3.02-um klin, Fri Nov  1 10:46:14 1991, APOLLO stuff added
 *      3.03-um klin, Tue Feb 11 22:58:03 1992, statusfile() into stat.c
 *                                              getopt() into sup/getopt.c
 *              klin, Fri Mar  6 10:45:49 1992, strclean() added
 *      3.04-um klin, Sat May 23 17:05:43 1992, Minor changes
 *            a klin, Mon Oct 19 09:36:38 1992, Stuff for german chars added
 *
 *      Copyright (c) 1991/92 Peter Klingebiel & UNIX Magazin Muenchen.
 *      For copying and distribution information see the file COPYRIGHT.
 */
#ifndef lint
static char sccsid[] = "@(#) utree 3.04a-um (klin) Oct 19 1992 util.c";
#endif  /* !lint */

#include "defs.h"

/* ---- Local variables and definitions ------------------------------- */

#define TRASH(c) ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '#')

/* ---- Functions and procedures -------------------------------------- */

#ifdef  USEGERMANCHARS

/*
 *      SPECIAL STUFF FOR GERMAN CHARACTERS
 */

/* German umlaute and map table for ISO and IBM encoding */
typedef struct  {
  unsigned char c;              /* German umlaut   */
  unsigned char g1;             /* Map character 1 */
  unsigned char g2;             /* Map character 2 */
} umlist;

LOCAL umlist umlaute[] = {
  { 0xc4, 'A', 'e' },           /* ISO Adieresis   */
  { 0xd6, 'O', 'e' },           /* ISO Odieresis   */
  { 0xdc, 'U', 'e' },           /* ISO Udieresis   */
  { 0xdf, 's', 's' },           /* ISO germandbls  */
  { 0xe4, 'a', 'e' },           /* ISO adieresis   */
  { 0xf6, 'o', 'e' },           /* ISO odieresis   */
  { 0xfc, 'u', 'e' },           /* ISO udieresis   */
  { 0x8e, 'A', 'e' },           /* IBM Adieresis   */
  { 0x99, 'O', 'e' },           /* IBM Odieresis   */
  { 0x9a, 'U', 'e' },           /* IBM Udieresis   */
  { 0xe1, 's', 's' },           /* IBM germandbls  */
  { 0x84, 'a', 'e' },           /* IBM adieresis   */
  { 0x94, 'o', 'e' },           /* IBM odieresis   */
  { 0x81, 'u', 'e' },           /* IBM udieresis   */
  { 0x00, '-', '-' },           /* End of table    */
};

#ifdef  __STDC__
# define CONST const
#else   /* !__STDC__ */
# define CONST register
#endif  /* __STDC__ */

/* Check if c is german umlaut */
GLOBL int isumlaut(c)
  register int c;
{
  register int i;

  for(i = 0; umlaute[i].c; i++)
    if(umlaute[i].c == (unsigned char) c)
      return(1);
  return(0);

} /* isumlaut */

/* Map german umlaut */
LOCAL VOID mapumlaut(c, g)
  register unsigned char *c, *g;
{
  register int i;

  for(i = 0, *g = '\0'; umlaute[i].c; i++)
    if(umlaute[i].c ==  *c) {
      *c = umlaute[i].g1;
      *g = umlaute[i].g2;
      break;
    }

} /* mapumlaut() */

/* Compare strings s1 and s2 with respect to german umlaute */
GLOBL int strcmp(s1, s2)
  CONST char *s1, *s2;
{
  unsigned char c1, c2, g1, g2;

  while(1) {
    c1 = (unsigned char) *s1++;
    c2 = (unsigned char) *s2++;
    if(c1 == c2) {
      if(c1)
	continue;
      else
	break;
    }
    if(isumlaut(c1) || isumlaut(c2)) {
      mapumlaut(&c1, &g1);
      mapumlaut(&c2, &g2);
      if(c1 == c2) {
	if(g1 == '\0')
	  g1 = (unsigned char) *s1++;
	if(g2 == '\0')
	  g2 = (unsigned char) *s2++;
	if(g1 == g2)
	  continue;
	else
	  return(g1 - g2);
      }
      else
	break;
    }
    else
      break;
  }
  return(c1 - c2);

} /* strcmp() */

/* Compare strings s1 and s2 up to length n with respect to german umlaute */
GLOBL int strncmp(s1, s2, n)
  CONST char *s1, *s2;
  register int n;
{
  unsigned char c1, c2, g1, g2;

  while(n-- > 0) {
    c1 = (unsigned char) *s1++;
    c2 = (unsigned char) *s2++;
    if(c1 == c2) {
      if(c1)
	continue;
      else
	break;
    }
    if(isumlaut(c1) || isumlaut(c2)) {
      mapumlaut(&c1, &g1);
      mapumlaut(&c2, &g2);
      if(c1 == c2) {
	if(g1 == '\0')
	  g1 = (unsigned char) *s1++;
	if(g2 == '\0')
	  g2 = (unsigned char) *s2++;
	if(g1 == g2)
	  continue;
	else
	  return(g1 - g2);
      }
      else
	break;
    }
    else
      break;
  }
  return(c1 - c2);

} /* strncmp() */
#endif  /* USEGERMANCHARS */

/*
 *      COMMON USED UTILITIY FUNCTIONS
 *
 */

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* !  This function is a modified version of adjustname() `stolen'    ! */
/* !  from the public domain editor MICRO GNU EMACS (mg) version 2a.  ! */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/* Return pathname of filename fn and current directory cwd */
GLOBL char *pathname(fn, cwd)
  register char *fn, *cwd;
{
  static char pn[NAMELEN];
  register char *cp;

  if(*fn == '/') {
    cp = pn;
#ifdef  APOLLO                  /* Stuff for APOLLO node names: //node */
    if(fn[1] == '/')
      *cp++ = *fn++;
#endif  /* APOLLO */
    while(fn[1] == '/')
      ++fn;
#ifdef  MUNET                   /* Stuff for MUNIX/NET: /../node */
    if(fn[1] == '.' && fn[2] == '.') {
      *cp++ = *fn++;
      *cp++ = *fn++;
    }
#endif  /* MUNET */
    *cp++ = *fn++;
  }
  else {
    (void) strcpy(pn, cwd);
    cp = pn + strlen(pn);
  }
  if(cp != pn && cp[-1] != '/')
    *cp++ = '/';
  while(*fn) {
    switch(*fn) {
      case '.':
	switch(fn[1]) {
	  case '\0':
	    *--cp = '\0';
	    return(pn);
	  case '/':
	    fn += 2;
	    continue;
	  case '.':
	    if(fn[2]=='/' || fn[2] == '\0') {
	      --cp;
	      while(cp > pn && *--cp != '/')
		;
	      ++cp;
	      if(fn[2]=='\0') {
		*--cp = '\0';
		return(pn);
	      }
	      fn += 3;
	      continue;
	    }
	    break;
	  default:
	    break;
	}
	break;
      case '/':
	fn++;
	continue;
      default:
	break;
    }
    while(*fn && (*cp++ = *fn++) != '/')
      ;
  }
#ifdef  APOLLO
  if(cp != &pn[2] && cp[-1] == '/')
#else   /* !APOLLO */
  if(cp != &pn[1] && cp[-1] == '/')
#endif  /* APOLLO */
    --cp;
  *cp = '\0';

  return(pn);

} /* pathname() */

/* Return basename of filename s */
GLOBL char *basename(s)
  register char *s;
{
  register char *sp;

  sp = s + strlen(s);
  while(sp >= s)
    if(*sp == '/')
      return(*(++sp) ? sp : s);
    else
      --sp;
  return(s);

} /* basename() */

/* Save string s */
GLOBL char *strsav(s)
  register char *s;
{
  register char *sp;

  if(sp = ualloc((unsigned) strlen(s) + 1, sizeof(char)))
    (void) strcpy(sp, s);
  return(sp);

} /* strsav() */

/* Convert string from lower to upper case */
GLOBL VOID strupper(s)
  register char *s;
{
  while(*s) {
    if(islower(*s))
      *s = toupper(*s);
    ++s;
  }

} /* strupper() */

/* Delete leading and trailing trash characters from s */
GLOBL char *strclean(s)
  register char *s;
{
  register char *p;

  while(*s && TRASH(*s))
    ++s;
  p = s + strlen(s);
  while(p > s && TRASH(*(p-1)))
    --p;
  *p = '\0';
  return(s);

} /* strclean() */

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* !   This function is a modified and tuned version of a wild card   ! */
/* !   pattern matcher written by Rich Salz (mirror!rs, Nov 26 1986)  ! */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/* Shell like pattern matching for meta characters *, ?, [class] */
/* Check string s against pattern p and return 1 on match 0 else */
GLOBL int match(s, p)
  register char *s, *p;
{
  register int lc, ok, ex;

  for( ; *p; s++, p++) {
    switch(*p) {
      case '\\':                /* Literal match next char */
	++p;
	/*FALLTHRU*/
      default:                  /* Literal match char */
	if(*s != *p)
	  return(0);
	continue;
      case '?':                 /* Match any char */
	if(*s == '\0')
	  return(0);
	continue;
      case '*':                 /* Match any chars */
	if(*++p == '\0')        /* Matches all */
	  return(1);
	for( ; *s; s++)
	  if(match(s, p))
	    return(1);
	return(0);
      case '[':                 /* Class */
	if(ex = (p[1] == '^' || p[1] == '!'))
	  ++p;
	for(lc = 0400, ok = 0; *++p && *p != ']'; lc = *p)
	  if(*p == '-' ? *s <= *++p && *s >= lc : *s == *p)
	    ok = 1;
	if(ok == ex)
	  return(0);
	continue;
    }
  }
  return(*s == '\0');

} /* match() */

/* Get and return next valid filename entry from directory file d */
GLOBL char *readdname(dp)
  register DIR *dp;
{
#if     defined(BSD) || !defined(NODIRENT)
#ifdef  BSD
  register struct direct *d;
#else   /* SYSV */
  register struct dirent *d;
#endif  /* BSD */
  register char *f;

  /* Get next valid directory entry and return filename */
  while(d = readdir(dp)) {
    f = d->d_name;
    /* Skip "." and ".." */
    if(f[0] == '.' && (f[1] == '\0' || (f[1] == '.' && f[2] == '\0')))
      continue;
    return(f);
#else   /* SYSV && NODIRENT */
  static char n[DIRSIZ+1];
  struct direct d;
  register char *f;

  while(fread(&d, sizeof(struct direct), 1, dp) > 0) {
    f = d.d_name;
    /* Skip removed files, "." and ".." */
    if(d.d_ino == 0 || (f[0] == '.' && (f[1] == '\0' || (f[1] == '.' && f[2] == '\0'))))
      continue;
    (void) strncpy(n, d.d_name, DIRSIZ);
    n[DIRSIZ] = '\0';
    return(n);
#endif  /* BSD || !NODIRENT */
  }
  return(NULL);

} /* readdname() */
